#pragma once

#define FASTSCAPELIB_VERSION_MAJOR "0"
#define FASTSCAPELIB_VERSION_MINOR "1"
#define FASTSCAPELIB_VERSION_PATCH "2"

#define FASTSCAPELIB_VERSION "0.1.2"

#define FASTSCAPELIB_GIT_HASH_FULL "NOHASH"

#include <string>


namespace fastscapelib
{

namespace version
{

    static const std::string version_major = FASTSCAPELIB_VERSION_MAJOR;
    static const std::string version_minor = FASTSCAPELIB_VERSION_MINOR;
    static const std::string version_patch = FASTSCAPELIB_VERSION_PATCH;

    static const std::string version = FASTSCAPELIB_VERSION;
    static const std::string git_hash_full = FASTSCAPELIB_GIT_HASH_FULL;

}  // namespace version

}  // namespace fastscapelib
